@@ -13,45 +13,26 @@
  * proven.  In reality, I don't believe it'll make one darn bit of
  * difference.  USB v1.1 is a slow bus by today's standards.
  *
+ * OHCI hardware takes care of most of the scheduling of different
+ * transfer types with the correct prioritization for us.
+ *
  * To get started in USB, I used the "Universal Serial Bus System
  * Architecture" book by Mindshare, Inc.  It was a reasonable introduction
- * and overview of USB and the two dominate host controller interfaces
- * however you're probably better off just reading the real specs
- * available from www.usb.org as you'll need them to get enough
- * details to actually implement a HCD.  The book has many typos and
- * omissions.  Beware, the specs are the victim of a committee.
+ * and overview of USB and the two dominant host controller interfaces
+ * however you're better off just reading the real specs available
+ * from www.usb.org as you'll need them to get enough detailt to
+ * actually implement a HCD.  The book has many typos and omissions
+ * Beware, the specs are the victim of a committee.
  *
  * This code was written with Guinness on the brain, xsnow on the desktop
  * and Orbital, Orb, Enya & Massive Attack on the CD player.  What a life!  ;) 
  *
- * $Log: ohci.c,v $
- * Revision 1.5  1999/04/22 09:43:20  greg
- * This contains Vojtech Pavlik's 0.05 patches that merge in the JE tree
- * as well as converted some of my OHCI code towards the new more
- * generally useful system being used on the functioning UHCI code.
- * Thanks Vojtech, that saved me a lot of time! :)
- *
- * A first attempt at keyboard support as well as JE's hub code have been
- * rolled in.
- *
- * Revision 1.4  1999/04/21 07:15:54  greg
- * some more updates.
- *
- * Revision 1.3  1999/04/21 05:24:40  greg
- * Applied Bradley's usb-qh-possible-fix patch.  Also included some more
- * of my trudging along on OHCI support [no, its not done yet].
- *
- * Revision 1.2  1999/04/09 08:00:52  greg
- * Still plugging away on the OHCI support..  I'm putting this one up on
- * suitenine as 0.03b.
- *
- * Revision 1.1  1999/04/05 08:32:30  greg
- * Initial check in.  It compiles, but several key root hub functions
- * remain to be implemented as well as the interrupt function.  I'm
- * releasing this as is for now until I "finish" the rest.
+ * No filesystems were harmed in the development of this code.
  *
+ * $Id: ohci.c,v 1.11 1999/04/25 00:18:52 greg Exp $
  */
 
+#include <linux/config.h>
 #include <linux/module.h>
 #include <linux/pci.h>
 #include <linux/kernel.h>
@@ -70,20 +51,25 @@
 #include "ohci.h"
 #include "inits.h"
 
+#ifdef CONFIG_APM
+#include <linux/apm_bios.h>
+static int handle_apm_event(apm_event_t event);
+static int apm_resume = 0;
+#endif
+
 static struct wait_queue *ohci_configure = NULL;
 
 
-/*
- * FIXME
- * This function is also in uhci.c!  How about combining it into a
- * common HCI functions library once that is written (hci.c?)
- */
-static void wait_ms(unsigned int ms)
+static int ohci_td_result(struct ohci_device *dev, struct ohci_td *td)
 {
-       current->state = TASK_UNINTERRUPTIBLE;
-       schedule_timeout(1 + ms / 10);
-} /* wait_ms */
+	unsigned int status;
 
+	status = td->info & OHCI_TD_CC;
+
+	/* TODO Debugging code for TD failures goes here */
+
+	return status;
+}
 
 
 static spinlock_t ohci_edtd_lock = SPIN_LOCK_UNLOCKED;
@@ -95,24 +81,24 @@ static spinlock_t ohci_edtd_lock = SPIN_LOCK_UNLOCKED;
  */
 static void ohci_add_td_to_ed(struct ohci_td *td, struct ohci_ed *ed)
 {
-	struct ohci_td *tail = ed->tail_td;
-	struct ohci_td *head = ed->head_td;
+	struct ohci_td *tail = bus_to_virt(ed->tail_td);
+	struct ohci_td *head = bus_to_virt(ed->head_td);
 	unsigned long flags;
 
 	spin_lock_irqsave(&ohci_edtd_lock, flags);
 
 	if (tail == head) {	/* empty list, put it on the head */
-		head = virt_to_bus(td);
+		head = (struct ohci_td *) virt_to_bus(td);
 		tail = 0;
 	} else {
 		if (!tail) {	/* no tail, single element list */
-			td->td_next = head->td->td_next;
-			head->td->td_next = virt_to_bus(td);
-			tail = virt_to_bus(td);
+			td->next_td = head->next_td;
+			head->next_td = virt_to_bus(td);
+			tail = (struct ohci_td *) virt_to_bus(td);
 		} else {	/* append to the list */
-			td->td_next = tail->td_next;
-			tail->td_next = virt_to_bus(td);
-			tail = virt_to_bus(td);
+			td->next_td = tail->next_td;
+			tail->next_td = virt_to_bus(td);
+			tail = (struct ohci_td *) virt_to_bus(td);
 		}
 	}
 	/* save the reverse link */
@@ -127,8 +113,7 @@ static void ohci_add_td_to_ed(struct ohci_td *td, struct ohci_ed *ed)
  */
 static void ohci_remove_td_from_ed(struct ohci_td *td, struct ohci_ed *ed)
 {
-	struct ohci_td *tail = ed->tail_td;
-	struct ohci_td *head = ed->head_td;
+	struct ohci_td *head = bus_to_virt(ed->head_td);
 	struct ohci_td *tmp_td;
 	unsigned long flags;
 
@@ -145,7 +130,7 @@ static void ohci_remove_td_from_ed(struct ohci_td *td, struct ohci_ed *ed)
 	}
 
 	tmp_td = head;
-	head = ed->head_td;
+	head = (struct ohci_td *) ed->head_td;
 	
 	while (head != NULL) {
 
@@ -158,9 +143,9 @@ static void ohci_remove_td_from_ed(struct ohci_td *td, struct ohci_ed *ed)
 		head = bus_to_virt(head->next_td);
 	}
 
-	td->next_td = NULL;  /* remove links to ED list */
+	td->next_td = virt_to_bus(NULL);  /* remove links to ED list */
 
-	/* TODO XXX mark this TD for possible cleanup? */
+	/* XXX mark this TD for possible cleanup? */
 
 	/* unset the "skip me bit" in this ED */
 	writel_mask(~(__u32)OHCI_ED_SKIP, ed->status);
@@ -180,11 +165,11 @@ static struct ohci_td *ohci_get_free_td(struct ohci_device *dev)
 	int idx;
 
 	for (idx=0; idx < NUM_TDS; idx++) {
-		if (td_done(dev->td[idx])) {
+		if (td_done(dev->td[idx].info)) {
 			/* XXX should this also zero out the structure? */
 			/* mark all new TDs as unaccessed */
-			dev->td[idx]->info = OHCI_TD_CC_NEW;
-			return dev->td[idx];
+			dev->td[idx].info = OHCI_TD_CC_NEW;
+			return &dev->td[idx];
 		}
 	}
 
@@ -197,17 +182,16 @@ static struct ohci_td *ohci_get_free_td(struct ohci_device *dev)
  **********************************/
 static spinlock_t irqlist_lock = SPIN_LOCK_UNLOCKED;
 
-static void ohci_add_irq_list(struct ohci *ohci, struct ohci_td *td, usb_device_irq completed, void *data, void *dev_id)
+static void ohci_add_irq_list(struct ohci *ohci, struct ohci_td *td, usb_device_irq completed, void *dev_id)
 {
 	unsigned long flags;
 
 	/* save the irq in our private portion of the TD */
 	td->completed = completed;
-	td->data = data;
 	td->dev_id = dev_id;
 
 	spin_lock_irqsave(&irqlist_lock, flags);
-	list_add(&td->irq_list, &ohci->interrupt_list)
+	list_add(&td->irq_list, &ohci->interrupt_list);
 	spin_unlock_irqrestore(&irqlist_lock, flags);
 } /* ohci_add_irq_list() */
 
@@ -231,7 +215,7 @@ static void ohci_remove_irq_list(struct ohci_td *td)
  * match will be used.  Powers of two from 1-32 are supported by OHCI.
  */
 static int ohci_request_irq(struct usb_device *usb, unsigned int pipe,
-	usb_device_irq handler, void *data, int period, void *dev_id)
+	usb_device_irq handler, int period, void *dev_id)
 {
 	struct ohci_device *dev = usb_to_ohci(usb);
 	struct ohci_td *td = dev->td;	/* */
@@ -276,10 +260,10 @@ static int ohci_request_irq(struct usb_device *usb, unsigned int pipe,
 	 * A pointer to the last *byte* in the buffer (ergh.. we get
 	 * to work around C's pointer arithmatic here with a typecast)
 	 */
-	td->buf_end = virt_to_bus(((*u8)(dev->data + DATA_BUF_LEN)) - 1);
+	td->buf_end = virt_to_bus(((u8*)(dev->data + DATA_BUF_LEN)) - 1);
 
 	/* does this make sense for ohci?.. time to think.. */
-	ohci_add_irq_list(dev->ohci, td, handler, data, dev_id);
+	ohci_add_irq_list(dev->ohci, td, handler, dev_id);
 	wmb();     /* found in asm/system.h; scary concept... */
 	ohci_add_td_to_ed(td, interrupt_ed);
 
@@ -291,7 +275,7 @@ static int ohci_request_irq(struct usb_device *usb, unsigned int pipe,
  */
 static struct wait_queue *control_wakeup;
 
-static int ohci_control_completed(int stats, void *data, void *buffer, void *dev_id)
+static int ohci_control_completed(int stats, void *buffer, void *dev_id)
 {
 	wake_up(&control_wakeup);
 	return 0;
@@ -305,12 +289,12 @@ static int ohci_control_completed(int stats, void *data, void *buffer, void *dev
 static int ohci_run_control(struct ohci_device *dev, struct ohci_td *status_td)
 {
 	struct wait_queue wait = { current, NULL };
-	struct ohci_ed *control_ed = dev->ohci->root_hub->ed[ED_CONTROL];
+	struct ohci_ed *control_ed = &dev->ohci->root_hub->ed[ED_CONTROL];
 
-	current->stats = TASK_UNINTERRUPTIBLE;
+	current->state = TASK_UNINTERRUPTIBLE;
 	add_wait_queue(&control_wakeup, &wait);
 
-	ohci_add_irq_list(dev->ohci, status_td, ohci_control_completed);
+	ohci_add_irq_list(dev->ohci, status_td, ohci_control_completed, NULL);
 	ohci_add_td_to_ed(status_td, control_ed);
 
 	/* FIXME? isn't this a little gross */
@@ -345,12 +329,11 @@ static int ohci_control_msg(struct usb_device *usb, unsigned int pipe, void *cmd
 	 * device.  (but who wants to use a keyboard AND a mouse
 	 * anyways? ;)
 	 */
-	struct ohci_ed *control_ed = dev->ohci->root_hub->ed[ED_CONTROL];
+	struct ohci_ed *control_ed = &dev->ohci->root_hub->ed[ED_CONTROL];
 	struct ohci_td *control_td;
 	struct ohci_td *data_td;
-	struct ohci_td *final_td;
+	struct ohci_td *last_td;
 	__u32 data_td_info;
-	unsigned long dest, status;
 
 	/*
 	 * Set the max packet size, device speed, endpoint number, usb
@@ -368,7 +351,7 @@ static int ohci_control_msg(struct usb_device *usb, unsigned int pipe, void *cmd
 	 */
 
 	/* get a TD to send this control message with */
-	control_td = virt_to_bus(ohci_get_free_td(dev));
+	control_td = ohci_get_free_td(dev);
 	/* TODO check for NULL */
 
 	/*
@@ -405,7 +388,7 @@ static int ohci_control_msg(struct usb_device *usb, unsigned int pipe, void *cmd
 		/* point to the current spot in the data buffer */
 		data_td->cur_buf = virt_to_bus(data);
 		/* point to the end of this data */
-		data_td->buf_end = virt_to_bus(data+pktsize-1)
+		data_td->buf_end = virt_to_bus(data+pktsize-1);
 
 		/* allocate the next TD */
 		tmp_td = ohci_get_free_td(dev);  /* TODO check for NULL */
@@ -529,6 +512,7 @@ static void start_hc(struct ohci *ohci)
 static void ohci_reset_port(struct ohci *ohci, unsigned int port)
 {
 	short ms;
+	int status;
 
 	/* Don't allow overflows. */
 	if (port >= MAX_ROOT_PORTS) {
@@ -540,14 +524,23 @@ static void ohci_reset_port(struct ohci *ohci, unsigned int port)
 
 	/*
 	 * Get the time required for a root hub port to reset and wait
-	 * it out (adding 3ms for good measure).
+	 * it out (adding 1ms for good measure).
 	 */
-	ms = (readl(&ohci->regs->roothub.a) >> 24) * 2 + 3;
+	ms = (readl(&ohci->regs->roothub.a) >> 24) * 2 + 1;
 	wait_ms(ms);
 
-	/* TODO check port status to see that the reset completed */
+	/* check port status to see that the reset completed */
+	status = readl(&ohci->regs->roothub.portstatus[port]);
+	if (status & PORT_PRS) {
+		/* reset failed, try harder? */
+		printk("usb-ohci: port %d reset failed, retrying\n", port);
+		writel(PORT_PRS, &ohci->regs->roothub.portstatus[port]);
+		wait_ms(50);
+	}
+
+	/* TODO we might need to re-enable the port here or is that
+	 * done elsewhere? */
 
-	/* TODO we might need to re-enable the port? */
 } /* ohci_reset_port */
 
 
@@ -556,6 +549,7 @@ static void ohci_reset_port(struct ohci *ohci, unsigned int port)
  */
 static void ohci_connect_change(struct ohci * ohci, int port)
 {
+	struct usb_device *usb_dev;
 	struct ohci_device *dev;
 	/* memory I/O address of the port status register */
 	void *portaddr = &ohci->regs->roothub.portstatus[port];
@@ -566,12 +560,12 @@ static void ohci_connect_change(struct ohci * ohci, int port)
 	 * everything we think we know about the device
 	 * on this root hub port.  It may have changed.
 	 */
-	usb_disconnect(ohci->root_hub->usb.children + port);
+	usb_disconnect(ohci->root_hub->usb->children + port);
 
 	portstatus = readl(portaddr);
 
 	/* disable the port if nothing is connected */
-	if (!(status & PORT_CCS)) {
+	if (!(portstatus & PORT_CCS)) {
 		writel(PORT_CCS, portaddr);
 		return;
 	}
@@ -579,35 +573,29 @@ static void ohci_connect_change(struct ohci * ohci, int port)
 	/*
 	 * Allocate a device for the new thingy that's been attached
 	 */
-	dev = (struct ohci_device *) __get_free_page(GFP_KERNEL);
-	if (!dev) {
-	    printk("usb-ohci: device memory allocation failed");
-	    return;
-	}
+	usb_dev = ohci_usb_allocate(ohci->root_hub->usb);
+	dev = usb_dev->hcpriv;
 
-	/*
-	 * Initialize the device structure
-	 */
-	memset(dev, 0, sizeof(*dev));
 	dev->ohci = ohci;
-	usb_connect(&dev->usb, &ohci->devmap);
-	dev->usb.op = &ohci_device_operations;
+
+	usb_connect(dev->usb);
 
 	/* link it into the bus's device tree */
-	ohci->root_hub->usb.children[port] = &dev->usb;
+	ohci->root_hub->usb->children[port] = usb_dev;
 
-	ohci_reset_port(port);
+	wait_ms(200); /* wait for powerup; XXX is this needed? */
+	ohci_reset_port(ohci, port);
 
-	/* Get speed information */
-	dev->usb.slow = readl(portaddr) & PORT_LSDA ? 1 : 0;
+	/* Get information on speed by using LSD */
+	usb_dev->slow = readl(portaddr) & PORT_LSDA ? 1 : 0;
 
 	/*
 	 * Do generic USB device tree processing on the new device.
 	 */
-	usb_new_device(&dev->usb);
-
+	usb_new_device(usb_dev);
 } /* ohci_connect_change() */
 
+
 /*
  * This gets called when the root hub configuration
  * has changed.  Just go through each port, seeing if
@@ -625,104 +613,147 @@ static void ohci_check_configuration(struct ohci *ohci)
 } /* ohci_check_configuration() */
 
 
-
+/*
+ * Get annoyed at the controller for bothering us.
+ */
 static void ohci_interrupt(int irq, void *__ohci, struct pt_regs *r)
 {
 	struct ohci *ohci = __ohci;
 	struct ohci_regs *regs = ohci->regs;
+	struct ohci_hcca *hcca = ohci->root_hub->hcca;
+	__u32 donehead = hcca->donehead;
 
 	/*
-	 * Read the interrupt status and tell HC that we're ready for
-	 * another interrupt.   [TODO]
+	 * Check the interrupt status register if needed
 	 */
+	if (!donehead || (donehead & 1)) {
+		__u32 intrstatus = readl(&regs->intrstatus);
+
+		/*
+		 * XXX eek! printk's in an interrupt handler.  shoot me!
+		 */
+		if (intrstatus & OHCI_INTR_SO) {
+			printk(KERN_DEBUG "usb-ohci: scheduling overrun\n");
+		}
+		if (intrstatus & OHCI_INTR_RD) {
+			printk(KERN_DEBUG "usb-ohci: resume detected\n");
+		}
+		if (intrstatus & OHCI_INTR_UE) {
+			printk(KERN_DEBUG "usb-ohci: unrecoverable error\n");
+		}
+		if (intrstatus & OHCI_INTR_OC) {
+			printk(KERN_DEBUG "usb-ohci: ownership change?\n");
+		}
+
+		if (intrstatus & OHCI_INTR_RHSC) {
+			/* TODO Process events on the root hub */
+		}
+	}
 
 	/*
-	 * See which TD's completed.. [TODO]
+	 * Process the done list
 	 */
+	if (donehead &= ~0x1) {
+		/*
+		 * TODO See which TD's completed..
+		 */
+	}
 
-	/* Check if there are any events on the root hub [TODO] */
+	/* Re-enable done queue interrupts and reset the donehead */
+	hcca->donehead = 0;
+	writel(OHCI_INTR_WDH, &regs->intrenable);
 	
 } /* ohci_interrupt() */
 
 
-/*
- * This is a -simple- driver, in that spirit we'll statically allocate
- * a single ohci structure here.  Dynamic allocation can be tackled
- * later when someone shows me a good reason to support more than one
- * host controller at a time.
- */
-struct ohci static_ohci = {
-	0,		/* IRQ - filled in by PCI probe */
-	NULL,		/* OHCI Memory mapped I/O base address (virtual) */
-	{ { 0, } },	/* Device number bitmaps */
-};
-
 /*
  * Allocate the resources required for running an OHCI controller.
  * Host controller interrupts must not be running while calling this
- * function.
+ * function or the penguins will get angry.
  *
  * The mem_base parameter must be the usable -virtual- address of the
  * host controller's memory mapped I/O registers.
- *
- * This is where OHCI triumphs over UHCI, because good is dumb.
- * Note how much simpler this function is than in uhci.c.
- *
- * OHCI hardware takes care of most of the scheduling of different
- * transfer types with the correct prioritization for us.
  */
 static struct ohci *alloc_ohci(void* mem_base)
 {
 	int i;
-	struct ohci *ohci = &static_ohci;
-	struct ohci_device *hub;
+	struct ohci *ohci;
+	struct usb_bus *bus;
+	struct ohci_device *dev;
+	struct usb_device *usb;
+
+	ohci = kmalloc(sizeof(*ohci), GFP_KERNEL);
+	if (!ohci)
+		return NULL;
+
+	memset(ohci, 0, sizeof(*ohci));
 
 	ohci->irq = -1;
 	ohci->regs = mem_base;
 	INIT_LIST_HEAD(&ohci->interrupt_list);
 
+	bus = kmalloc(sizeof(*bus), GFP_KERNEL);
+	if (!bus)
+		return NULL;
+
+	memset(bus, 0, sizeof(*bus));
+
+	ohci->bus = bus;
+	bus->hcpriv = ohci;
+	bus->op = &ohci_device_operations;
+
 	/*
 	 * Here we allocate our own root hub and TDs as well as the
 	 * OHCI host controller communications area.  The HCCA is just
 	 * a nice pool of memory with pointers to endpoint descriptors
 	 * for the different interrupts.
-	 *
-	 * The first page of memory is our own structure and the TDs.
-	 * The second page of memory contains the HCCA.
 	 */
-	hub = (struct ohci_device *) __get_free_pages(GFP_KERNEL, 1);
-	ohci->root_hub = hub;
-	if (!hub)
+	usb = ohci_usb_allocate(NULL);
+	if (!usb)
 		return NULL;
 
-	/* Assign the next page of memory to the HCCA */
-	hub->hcca = (struct ohci_hcca *) (PAGE_SIZE + (unsigned long) hub);
+	dev = ohci->root_hub = usb_to_ohci(usb);
 
-	/* Tell the controller where the HCCA is */
-	writel(virt_to_bus(hub->hcca), &ohci->regs->hcca);
+	usb->bus = bus;
 
 	/* Initialize the root hub */
-	memset(hub, 0, sizeof(*hub));
-	hub->ohci = ohci;    /* link back to the controller */
+	memset(dev, 0, sizeof(*dev));
+	dev->ohci = ohci;    /* link back to the controller */
 
-	/* Get the number of downstream ports on this controller */
-	hub->usb.maxchild = readl(&ohci->regs->roothub.a) & 0xff;
-	if (hub->usb.maxchild > MAX_ROOT_PORTS) 
-		hub->usb.maxchild = MAX_ROOT_PORTS;
-	if (hub->usb.maxchild < 1) {
-		printk("Less than one root USB hub port? Impossible!\n");
-		hub->usb.maxchild = 1;
+	/*
+	 * Allocate the Host Controller Communications Area
+	 */
+	dev->hcca = (struct ohci_hcca *) kmalloc(sizeof(*dev->hcca), GFP_KERNEL);
+
+	/* Tell the controller where the HCCA is */
+	writel(virt_to_bus(dev->hcca), &ohci->regs->hcca);
+
+	/* Get the number of ports on the root hub */
+	usb->maxchild = readl(&ohci->regs->roothub.a) & 0xff;
+	if (usb->maxchild > MAX_ROOT_PORTS) {
+		printk("usb-ohci: Limited to %d ports\n", MAX_ROOT_PORTS);
+		usb->maxchild = MAX_ROOT_PORTS;
 	}
+	if (usb->maxchild < 1) {
+		printk("usb-ohci: Less than one root hub port? Impossible!\n");
+		usb->maxchild = 1;
+	}
+	printk("usb-ohci: %d root hub ports found\n", usb->maxchild);
+
+	printk("alloc_ohci() controller\n");
+	show_ohci_status(ohci);
+	printk("alloc_ohci() root_hub device\n");
+	show_ohci_device(dev);
 
 	/*
 	 * Initialize the ED polling "tree" (for simplicity's sake in
 	 * this driver many nodes in the tree will be identical)
 	 */
-	hub->ed[ED_INT_32].next_ed = virt_to_bus(&hub->ed[ED_INT_16]);
-	hub->ed[ED_INT_16].next_ed = virt_to_bus(&hub->ed[ED_INT_8]);
-	hub->ed[ED_INT_8].next_ed = virt_to_bus(&hub->ed[ED_INT_4]);
-	hub->ed[ED_INT_4].next_ed = virt_to_bus(&hub->ed[ED_INT_2]);
-	hub->ed[ED_INT_2].next_ed = virt_to_bus(&hub->ed[ED_INT_1]);
+	dev->ed[ED_INT_32].next_ed = virt_to_bus(&dev->ed[ED_INT_16]);
+	dev->ed[ED_INT_16].next_ed = virt_to_bus(&dev->ed[ED_INT_8]);
+	dev->ed[ED_INT_8].next_ed = virt_to_bus(&dev->ed[ED_INT_4]);
+	dev->ed[ED_INT_4].next_ed = virt_to_bus(&dev->ed[ED_INT_2]);
+	dev->ed[ED_INT_2].next_ed = virt_to_bus(&dev->ed[ED_INT_1]);
 
 	/*
 	 * Initialize the polling table to call interrupts at the
@@ -730,30 +761,30 @@ static struct ohci *alloc_ohci(void* mem_base)
 	 */
 	for (i = 0; i < NUM_INTS; i++) {
 		if (i == 0)
-			hub->hcca->int_table[i] =
-				virt_to_bus(&hub->ed[ED_INT_32]);
+			dev->hcca->int_table[i] =
+				virt_to_bus(&dev->ed[ED_INT_32]);
 		else if (i & 1)
-			hub->hcca->int_table[i] =
-				virt_to_bus(&hub->ed[ED_INT_16]);
+			dev->hcca->int_table[i] =
+				virt_to_bus(&dev->ed[ED_INT_16]);
 		else if (i & 2)
-			hub->hcca->int_table[i] =
-				virt_to_bus(&hub->ed[ED_INT_8]);
+			dev->hcca->int_table[i] =
+				virt_to_bus(&dev->ed[ED_INT_8]);
 		else if (i & 4)
-			hub->hcca->int_table[i] =
-				virt_to_bus(&hub->ed[ED_INT_4]);
+			dev->hcca->int_table[i] =
+				virt_to_bus(&dev->ed[ED_INT_4]);
 		else if (i & 8)
-			hub->hcca->int_table[i] =
-				virt_to_bus(&hub->ed[ED_INT_2]);
+			dev->hcca->int_table[i] =
+				virt_to_bus(&dev->ed[ED_INT_2]);
 		else if (i & 16)
-			hub->hcca->int_table[i] =
-				virt_to_bus(&hub->ed[ED_INT_1]);
+			dev->hcca->int_table[i] =
+				virt_to_bus(&dev->ed[ED_INT_1]);
 	}
 
 	/*
 	 * Tell the controller where the control and bulk lists are
 	 */
-	writel(virt_to_bus(&hub->ed[ED_CONTROL]), &ohci->regs->ed_controlhead);
-	writel(virt_to_bus(&hub->ed[ED_BULK]), &ohci->regs->ed_bulkhead);
+	writel(virt_to_bus(&dev->ed[ED_CONTROL]), &ohci->regs->ed_controlhead);
+	writel(virt_to_bus(&dev->ed[ED_BULK]), &ohci->regs->ed_bulkhead);
 
 	return ohci;
 } /* alloc_ohci() */
@@ -794,8 +825,7 @@ static int ohci_control_thread(void * __ohci)
 	
 	/*
 	 * I'm unfamiliar with the SMP kernel locking.. where should
-	 * this be released?  Or is this here to prevent SMP systems
-	 * from using more than 1 CPU at all for now... ?
+	 * this be released?  -greg
 	 */
 	lock_kernel();
 
@@ -816,6 +846,13 @@ static int ohci_control_thread(void * __ohci)
 	start_hc(ohci);
 	do {
 		interruptible_sleep_on(&ohci_configure);
+#ifdef CONFIG_APM
+		if (apm_resume) {
+			apm_resume = 0;
+			start_hc(ohci);
+			continue;
+		}
+#endif
 		ohci_check_configuration(ohci);
 	} while (!signal_pending(current));
 
@@ -826,6 +863,38 @@ static int ohci_control_thread(void * __ohci)
 	return 0;
 } /* ohci_control_thread() */
 
+
+#ifdef CONFIG_APM
+static int handle_apm_event(apm_event_t event)
+{
+	static int down = 0;
+
+	switch (event) {
+	case APM_SYS_SUSPEND:
+	case APM_USER_SUSPEND:
+		if (down) {
+			printk(KERN_DEBUG "usb-ohci: received extra suspend event\n");
+			break;
+		}
+		down = 1;
+		break;
+	case APM_NORMAL_RESUME:
+	case APM_CRITICAL_RESUME:
+		if (!down) {
+			printk(KERN_DEBUG "usb-ohci: received bogus resume event\n");
+			break;
+		}
+		down = 0;
+		if (waitqueue_active(&ohci_configure)) {
+			apm_resume = 1;
+			wake_up(&ohci_configure);
+		}
+		break;
+	}
+	return 0;
+}
+#endif
+
 /* ... */
 
 /*
@@ -842,17 +911,20 @@ static int found_ohci(int irq, void* mem_base)
 	if (!ohci)
 		return -ENOMEM;
 
+	printk("usb-ohci: alloc_ohci() = %p\n", ohci);
+
 	reset_hc(ohci);
 
 	retval = -EBUSY;
-	if (request_irq(irq, ohci_interrupt, SA_SHIRQ, "ohci-usb", ohci) == 0) {
+	if (request_irq(irq, ohci_interrupt, SA_SHIRQ, "usb-ohci", ohci) == 0) {
 		int pid;
 
 		MOD_INC_USE_COUNT;
 		ohci->irq = irq;
 
 		/* fork off the handler */
-		pid = kernel_thread(ohci_control_thread, ohci, CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
+		pid = kernel_thread(ohci_control_thread, ohci,
+				CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
 		if (pid >= 0)
 			return 0;
 
@@ -870,6 +942,8 @@ static int found_ohci(int irq, void* mem_base)
 static int init_ohci(struct pci_dev *dev)
 {
 	unsigned int mem_base = dev->base_address[0];
+	
+	printk("usb-ohci: mem_base is %p\n", (void*)mem_base);
 
 	/* If its OHCI, its memory */
 	if (mem_base & PCI_BASE_ADDRESS_SPACE_IO)
@@ -893,49 +967,68 @@ static int init_ohci(struct pci_dev *dev)
 	return found_ohci(dev->irq, (void *) mem_base);
 } /* init_ohci() */
 
-/* FIXME this should be named following Linux convention and go in pci.h */
-#define PCI_CLASS_SERIAL_USB_OHCI 0x10
+
+/* TODO this should be named following Linux convention and go in pci.h */
+#define PCI_CLASS_SERIAL_USB_OHCI ((PCI_CLASS_SERIAL_USB << 8) | 0x0010)
 
 /*
  * Search the PCI bus for an OHCI USB controller and set it up
  *
- * If anyone -really- wants multiple controllers this will need to be
+ * If anyone wants multiple controllers this will need to be
  * updated..  Right now, it just picks the first one it finds.
  */
 int init_module(void)
 {
 	int retval;
 	struct pci_dev *dev = NULL;
-	u8 type;
+	/*u8 type;*/
 
 	if (sizeof(struct ohci_device) > 4096) {
-		printk("OHCI Aborting: struct ohci_device to large\n");
+		printk("usb-ohci: struct ohci_device to large\n");
 		return -ENODEV;
 	}
 
 	retval = -ENODEV;
 	for (;;) {
-		/* Find a USB controller */
-		dev = pci_find_class(PCI_CLASS_SERIAL_USB<<8, dev);
+		/* Find an OHCI USB controller */
+		dev = pci_find_class(PCI_CLASS_SERIAL_USB_OHCI, dev);
 		if (!dev)
 			break;
-		/* It is OpenHCI? */
-		pci_read_config_byte(dev, PCI_CLASS_PROG, &type);
-		if (type!=PCI_CLASS_SERIAL_USB_OHCI)
-			continue;
+
+		/* Verify that its OpenHCI by checking for MMIO */
+		/* pci_read_config_byte(dev, PCI_CLASS_PROG, &type);
+		if (!type)
+			continue; */
+
 		/* Ok, set it up */
 		retval = init_ohci(dev);
 		if (retval < 0)
 			continue;
 
-		usb_mouse_init();
-		usb_kbd_init();
-		hub_init();
+		/* TODO check module params here to determine what to load */
+
+/*		usb_mouse_init(); */
+/*		usb_kbd_init();
+		hub_init();	*/
+#ifdef CONFIG_APM
+		apm_register_callback(&handle_apm_event);
+#endif
 
 		return 0; /* no error */
 	}
 	return retval;
 } /* init_module() */
 
+
+/*
+ *  Clean up when unloading the module
+ */
+void cleanup_module(void)
+{
+#ifdef CONFIG_APM
+	apm_unregister_callback(&handle_apm_event);
+#endif
+}
+
 /* vim:sw=8
  */
