@@ -20,6 +20,7 @@
 
 /* 4/4/1999 added data toggle for interrupt pipes -keryan */
 
+#include <linux/config.h>
 #include <linux/module.h>
 #include <linux/pci.h>
 #include <linux/kernel.h>
@@ -38,6 +39,12 @@
 #include "uhci.h"
 #include "inits.h"
 
+#ifdef CONFIG_APM
+#include <linux/apm_bios.h>
+static int handle_apm_event(apm_event_t event);
+static int apm_resume = 0;
+#endif
+
 #define compile_assert(x) do { switch (0) { case 1: case !(x): } } while (0)
 
 int usb_mouse_init(void);
@@ -45,12 +52,6 @@ int hub_init(void);
 
 static struct wait_queue *uhci_configure = NULL;
 
-static void wait_ms(unsigned int ms)
-{
-	current->state = TASK_UNINTERRUPTIBLE;
-	schedule_timeout(1 + ms / 10);
-}
-
 /*
  * Return the result of a TD..
  */
@@ -117,6 +118,32 @@ static inline void uhci_insert_td_in_qh(struct uhci_qh *qh, struct uhci_td *td)
 	uhci_insert_tds_in_qh(qh, td, td);
 }
 
+static void uhci_insert_qh(struct uhci_qh *qh, struct uhci_qh *newqh)
+{
+	newqh->link = qh->link;
+	qh->link = virt_to_bus(newqh) | 2;
+}
+
+static void uhci_remove_qh(struct uhci_qh *qh, struct uhci_qh *remqh)
+{
+	unsigned int remphys = virt_to_bus(remqh);
+	struct uhci_qh *lqh = qh;
+
+	while ((lqh->link & ~0xF) != remphys) {
+		if (lqh->link & 1)
+			break;
+
+		lqh = bus_to_virt(lqh->link & ~0xF);
+	}
+
+	if (lqh->link & 1) {
+		printk("couldn't find qh in chain!\n");
+		return;
+	}
+
+	lqh->link = remqh->link;
+}
+
 /*
  * Removes td from qh if present.
  *
@@ -163,6 +190,60 @@ static void uhci_remove_td(struct uhci_td *td)
 		:"memory");
 }
 
+static struct uhci_qh *uhci_qh_allocate(struct uhci_device *dev)
+{
+	struct uhci_qh *qh;
+	int inuse;
+
+	qh = dev->qh;
+	for (; (inuse = test_and_set_bit(0, &qh->inuse)) != 0 && qh < &dev->qh[UHCI_MAXQH]; qh++)
+		;
+
+	if (!inuse)
+		return(qh);
+
+	printk("ran out of qh's for dev %p\n", dev);
+	return(NULL);
+}
+
+static void uhci_qh_deallocate(struct uhci_qh *qh)
+{
+	if (qh->element != 1)
+		printk("qh %p leaving dangling entries? (%X)\n", qh, qh->element);
+
+	qh->element = 1;
+	qh->link = 1;
+
+	clear_bit(0, &qh->inuse);
+}
+
+static struct uhci_td *uhci_td_allocate(struct uhci_device *dev)
+{
+	struct uhci_td *td;
+	int inuse;
+
+	td = dev->td;
+	for (; (inuse = test_and_set_bit(0, &td->inuse)) != 0 && td < &dev->td[UHCI_MAXTD]; td++)
+		;
+
+	if (!inuse)
+		return(td);
+
+	printk("ran out of td's for dev %p\n", dev);
+	return(NULL);
+}
+
+/*
+ * This MUST only be called when it has been removed from a QH already (or
+ * the QH has been removed from the skeleton
+ */
+static void uhci_td_deallocate(struct uhci_td *td)
+{
+	td->link = 1;
+
+	clear_bit(0, &td->inuse);
+}
+
 /*
  * UHCI interrupt list operations..
  */
@@ -192,11 +273,11 @@ static void uhci_remove_irq_list(struct uhci_td *td)
 /*
  * Request a interrupt handler..
  */
-static int uhci_request_irq(struct usb_device * usb_dev, unsigned int pipe, usb_device_irq handler, int period, void *dev_id)
+static int uhci_request_irq(struct usb_device *usb_dev, unsigned int pipe, usb_device_irq handler, int period, void *dev_id)
 {
 	struct uhci_device *dev = usb_to_uhci(usb_dev);
-	struct uhci_td *td = dev->td + 30;					/* cheesy hack */
-	struct uhci_qh *interrupt_qh = &dev->uhci->root_hub->interrupt8_qh;	/* Ignore the period for now */
+	struct uhci_td *td = uhci_td_allocate(dev);
+	struct uhci_qh *interrupt_qh = uhci_qh_allocate(dev);
 
 	unsigned int destination, status;
 
@@ -214,10 +295,15 @@ static int uhci_request_irq(struct usb_device * usb_dev, unsigned int pipe, usb_
 	td->status = status;			/* In */
 	td->info = destination | (7 << 21);	/* 8 bytes of data */
 	td->buffer = virt_to_bus(dev->data);
+	td->qh = interrupt_qh;
+	interrupt_qh->skel = &dev->uhci->root_hub->skel_int8_qh;
 
 	uhci_add_irq_list(dev->uhci, td, handler, dev_id);
 
 	uhci_insert_td_in_qh(interrupt_qh, td);
+
+	/* Add it into the skeleton */
+	uhci_insert_qh(&dev->uhci->root_hub->skel_int8_qh, interrupt_qh);
 	return 0;
 }
 
@@ -238,44 +324,58 @@ static int uhci_control_completed(int status, void *buffer, void *dev_id)
 }
 
 /* td points to the last td in the list, which interrupts on completion */
-static int uhci_run_control(struct uhci_device *dev, struct uhci_td *td)
+static int uhci_run_control(struct uhci_device *dev, struct uhci_td *first, struct uhci_td *last)
 {
 	struct wait_queue wait = { current, NULL };
-	struct uhci_qh *ctrl_qh;
+	struct uhci_qh *ctrl_qh = uhci_qh_allocate(dev);
+	struct uhci_td *curtd;
 
 	current->state = TASK_UNINTERRUPTIBLE;
 	add_wait_queue(&control_wakeup, &wait);
 
-	uhci_add_irq_list(dev->uhci, td, uhci_control_completed, NULL);
+	uhci_add_irq_list(dev->uhci, last, uhci_control_completed, NULL);
 	
-	ctrl_qh = &dev->uhci->root_hub->control_qh;
-	
-	if(ctrl_qh->element != 1)
-		printk("c->e = 0x%x    td = 0x%p\n", ctrl_qh->element, td);
+	/* FIXME: This is kinda kludged */
+	/* Walk the TD list and update the QH pointer */
+	{
+	int maxcount = 100;
 
-	uhci_insert_tds_in_qh(ctrl_qh, dev->control_td, td);
+	curtd = first;
+	do {
+		curtd->qh = ctrl_qh;
+		if (curtd->link & 1)
+			break;
+
+		curtd = bus_to_virt(curtd->link & ~0xF);
+		if (!--maxcount) {
+			printk("runaway tds!\n");
+			break;
+		}
+	} while (1);
+	}
+
+	uhci_insert_tds_in_qh(ctrl_qh, first, last);
+
+	/* Add it into the skeleton */
+	uhci_insert_qh(&dev->uhci->root_hub->skel_control_qh, ctrl_qh);
 
 	schedule_timeout(HZ/10);
 
 	remove_wait_queue(&control_wakeup, &wait);
 
 	/* Clean up in case it failed.. */
-	uhci_remove_irq_list(td);
+	uhci_remove_irq_list(last);
 
 #if 0
 	printk("Looking for tds [%p, %p]\n", dev->control_td, td);
 #endif
-	{
-		int maxcount = 100;
-		struct uhci_td *first = dev->control_td;
-		do {
-			uhci_remove_td(first);
-			first++;
-			if (!--maxcount) break;
-		} while (first <= td);
-	}
 
-	return uhci_td_result(dev, td);
+	/* Remove it from the skeleton */
+	uhci_remove_qh(&dev->uhci->root_hub->skel_control_qh, ctrl_qh);
+
+	uhci_qh_deallocate(ctrl_qh);
+
+	return uhci_td_result(dev, last);
 }
 
 /*
@@ -302,8 +402,14 @@ static int uhci_run_control(struct uhci_device *dev, struct uhci_td *td)
 static int uhci_control_msg(struct usb_device *usb_dev, unsigned int pipe, void *cmd, void *data, int len)
 {
 	struct uhci_device *dev = usb_to_uhci(usb_dev);
-	struct uhci_td *td = dev->control_td;
+	struct uhci_td *first, *td, *prevtd;
 	unsigned long destination, status;
+	int ret;
+
+	if (len > usb_maxpacket(usb_dev->maxpacketsize) * 29)
+		printk("Warning, too much data for a control packet, crashing\n");
+
+	first = td = uhci_td_allocate(dev);
 
 	/* The "pipe" thing contains the destination in bits 8--18, 0x2D is SETUP */
 	destination = (pipe & 0x0007ff00) | 0x2D;
@@ -314,7 +420,6 @@ static int uhci_control_msg(struct usb_device *usb_dev, unsigned int pipe, void
 	/*
 	 * Build the TD for the control request
 	 */
-	td->link = 4 | virt_to_bus(td+1);		/* Point to next TD for input data */
 	td->status = status;				/* Try forever */
 	td->info = destination | (7 << 21);		/* 8 bytes of data */
 	td->buffer = virt_to_bus(cmd);
@@ -327,7 +432,9 @@ static int uhci_control_msg(struct usb_device *usb_dev, unsigned int pipe, void
 	if (usb_pipeout(pipe))
 		destination ^= (0xE1 ^ 0x69);		/* IN -> OUT */
 
-	td++;
+	prevtd = td;
+	td = uhci_td_allocate(dev);
+	prevtd->link = 4 | virt_to_bus(td);
 
 	/*
 	 * Build the DATA TD's
@@ -343,13 +450,15 @@ static int uhci_control_msg(struct usb_device *usb_dev, unsigned int pipe, void
 		/* Alternate Data0/1 (start with Data1) */
 		destination ^= 1 << 19;
 	
-		td->link = 4 | virt_to_bus(td+1);			/* Point to next TD for data */
 		td->status = status;					/* Status */
 		td->info = destination | ((pktsze-1) << 21);		/* pktsze bytes of data */
 		td->buffer = virt_to_bus(data);
-		td->backptr = &td[-1].link;
+		td->backptr = &prevtd->link;
+
+		prevtd = td;
+		td = uhci_td_allocate(dev);
+		prevtd->link = 4 | virt_to_bus(td);			/* Update previous TD */
 
-		td++;
 		data += maxsze;
 		len -= maxsze;
 	}
@@ -364,16 +473,39 @@ static int uhci_control_msg(struct usb_device *usb_dev, unsigned int pipe, void
 	td->status = status | (1 << 24);		/* IOC */
 	td->info = destination | (0x7ff << 21);		/* 0 bytes of data */
 	td->buffer = 0;
-	td->backptr = &td[-1].link;
+	td->backptr = &prevtd->link;
 
 	/* Start it up.. */
-	return uhci_run_control(dev, td);
+	ret = uhci_run_control(dev, first, td);
+
+	{
+		int maxcount = 100;
+		struct uhci_td *curtd = first;
+		unsigned int nextlink;
+
+		do {
+			nextlink = curtd->link;
+			uhci_remove_td(curtd);
+			uhci_td_deallocate(curtd);
+			if (nextlink & 1)	/* Tail? */
+				break;
+
+			curtd = bus_to_virt(nextlink & ~0xF);
+			if (!--maxcount) {
+				printk("runaway td's!?\n");
+				break;
+			}
+		} while (1);
+	}
+
+	return ret;
 }
 
 static struct usb_device *uhci_usb_allocate(struct usb_device *parent)
 {
 	struct usb_device *usb_dev;
 	struct uhci_device *dev;
+	int i;
 
 	usb_dev = kmalloc(sizeof(*usb_dev), GFP_KERNEL);
 	if (!usb_dev)
@@ -400,6 +532,18 @@ static struct usb_device *uhci_usb_allocate(struct usb_device *parent)
 		dev->uhci = usb_to_uhci(parent)->uhci;
 	}
 
+	/* Reset the QH's and TD's */
+	for (i = 0; i < UHCI_MAXQH; i++) {
+		dev->qh[i].link = 1;
+		dev->qh[i].element = 1;
+		dev->qh[i].inuse = 0;
+	}
+
+	for (i = 0; i < UHCI_MAXTD; i++) {
+		dev->td[i].link = 1;
+		dev->td[i].inuse = 0;
+	}
+
 	return usb_dev;
 }
 
@@ -409,17 +553,23 @@ static int uhci_usb_deallocate(struct usb_device *usb_dev)
 	int i;
 
 	/* There are UHCI_MAXTD preallocated tds */
-	for(i = 0; i < UHCI_MAXTD; ++i) {
+	for (i = 0; i < UHCI_MAXTD; ++i) {
 		struct uhci_td *td = dev->td + i;
-		int j;
 
-		/* Remove the td from any queues */
-		for(j = 0; j < UHCI_MAXQH; ++j)
-			uhci_remove_td(td);
-		
 		/* And remove it from the irq list, if it's active */
-		if(td->status & (1 << 23))
+		if (td->status & (1 << 23))
 			uhci_remove_irq_list(td);
+
+		if (td->inuse)
+			uhci_remove_td(td);
+	}
+
+	/* Remove the td from any queues */
+	for (i = 0; i < UHCI_MAXQH; ++i) {
+		struct uhci_qh *qh = dev->qh + i;
+
+		if (qh->inuse)
+			uhci_remove_qh(qh->skel, qh);
 	}
 
 	kfree(dev);
@@ -561,14 +711,19 @@ static void uhci_interrupt_notify(struct uhci *uhci)
 			__list_del(tmp->prev, next);
 			INIT_LIST_HEAD(tmp);
 			if (td->completed(td->status, bus_to_virt(td->buffer), td->dev_id)) {
-				struct uhci_qh * interrupt_qh = &uhci->root_hub->interrupt8_qh;
+				struct uhci_qh *interrupt_qh = td->qh;
+
 				list_add(&td->irq_list, &uhci->interrupt_list);
 				td->info ^= 1 << 19; /* toggle between data0 and data1 */
 				td->status = (td->status & 0x2f000000) | (1 << 23) | (1 << 24);	/* active */
 
+				/* Remove then readd? Is that necessary */
 				uhci_remove_td(td);
 				uhci_insert_td_in_qh(interrupt_qh, td);
 			}
+			/* If completed wants to not reactivate, then it's */
+			/* responsible for free'ing the TD's and QH's */
+			/* or another function (such as run_control) */
 		}
 		tmp = next;
 	}
@@ -627,12 +782,14 @@ static void uhci_interrupt(int irq, void *__uhci, struct pt_regs *regs)
 static void uhci_init_ticktd(struct uhci *uhci)
 {
 	struct uhci_device *dev = uhci->root_hub;
-	struct uhci_td *td = &dev->tick_td;
+	struct uhci_td *td = uhci_td_allocate(dev);
 
 	td->link = 1;
 	td->status = (1 << 24);					/* interrupt on completion */
 	td->info = (15 << 21) | 0x7f69;				/* (ignored) input packet, 16 bytes, device 127 */
 	td->buffer = 0;
+	td->qh = NULL;
+
 	uhci->fl->frame[0] = virt_to_bus(td);
 }
 
@@ -712,7 +869,8 @@ static struct uhci *alloc_uhci(unsigned int io_addr)
 	uhci->io_addr = io_addr;
 	INIT_LIST_HEAD(&uhci->interrupt_list);
 
-	uhci->fl = (void *)__get_free_page(GFP_KERNEL);	/* FIXME: Kludge? */
+	/* We need exactly one page (per UHCI specs), how convenient */
+	uhci->fl = (void *)__get_free_page(GFP_KERNEL);
 
 	bus = kmalloc(sizeof(*bus), GFP_KERNEL);
 	if (!bus)
@@ -743,32 +901,35 @@ static struct uhci *alloc_uhci(unsigned int io_addr)
 	usb->bus = bus;
 
 	/* Initialize the root hub */
-	usb->maxchild = 2;	/* FIXME: Hard coded to 2? */
+	/* UHCI specs says devices must have 2 ports, but goes on to say */
+	/* they may have more but give no way to determine how many they */
+	/* have, so default to 2 */
+	usb->maxchild = 2;
 	usb_init_root_hub(usb);
 
 	/*
 	 * Initialize the queues. They all start out empty,
 	 * linked to each other in the proper order.
 	 */
-	for (i = 0 ; i < 8; i++) {
-		dev->qh[i].link = 2 | virt_to_bus(&dev->control_qh);
+	for (i = 1 ; i < 9; i++) {
+		dev->qh[i].link = 2 | virt_to_bus(&dev->skel_control_qh);
 		dev->qh[i].element = 1;
 	}
 	
-	dev->control_qh.link = 2 | virt_to_bus(&dev->bulk0_qh);
-	dev->control_qh.element = 1;
+	dev->skel_control_qh.link = 2 | virt_to_bus(&dev->skel_bulk0_qh);
+	dev->skel_control_qh.element = 1;
 
-	dev->bulk0_qh.link = 2 | virt_to_bus(&dev->bulk1_qh);
-	dev->bulk0_qh.element = 1;
+	dev->skel_bulk0_qh.link = 2 | virt_to_bus(&dev->skel_bulk1_qh);
+	dev->skel_bulk0_qh.element = 1;
 
-	dev->bulk1_qh.link = 2 | virt_to_bus(&dev->bulk2_qh);
-	dev->bulk1_qh.element = 1;
+	dev->skel_bulk1_qh.link = 2 | virt_to_bus(&dev->skel_bulk2_qh);
+	dev->skel_bulk1_qh.element = 1;
 
-	dev->bulk2_qh.link = 2 | virt_to_bus(&dev->bulk3_qh);
-	dev->bulk2_qh.element = 1;
+	dev->skel_bulk2_qh.link = 2 | virt_to_bus(&dev->skel_bulk3_qh);
+	dev->skel_bulk2_qh.element = 1;
 
-	dev->bulk3_qh.link = 1;
-	dev->bulk3_qh.element = 1;
+	dev->skel_bulk3_qh.link = 1;
+	dev->skel_bulk3_qh.element = 1;
 
 	/*
 	 * Fill the frame list: make all entries point to
@@ -779,7 +940,7 @@ static struct uhci *alloc_uhci(unsigned int io_addr)
 	 * us a reasonable dynamic range for irq latencies.
 	 */
 	for (i = 0; i < 1024; i++) {
-		struct uhci_qh * irq = &dev->interrupt2_qh;
+		struct uhci_qh * irq = &dev->skel_int2_qh;
 		if (i & 1) {
 			irq++;
 			if (i & 2) {
@@ -818,10 +979,12 @@ static void release_uhci(struct uhci *uhci)
 		uhci->irq = -1;
 	}
 
+#if 0
 	if (uhci->root_hub) {
 		uhci_usb_deallocate(uhci_to_usb(uhci->root_hub));
 		uhci->root_hub = NULL;
 	}
+#endif
 
 	if (uhci->fl) {
 		free_page((unsigned long)uhci->fl);
@@ -835,7 +998,6 @@ static void release_uhci(struct uhci *uhci)
 static int uhci_control_thread(void * __uhci)
 {
 	struct uhci *uhci = (struct uhci *)__uhci;
-	int i;
 
 	lock_kernel();
 	request_region(uhci->io_addr, 32, "usb-uhci");
@@ -860,6 +1022,13 @@ static int uhci_control_thread(void * __uhci)
 		int unsigned long signr;
 
 		interruptible_sleep_on(&uhci_configure);
+#ifdef CONFIG_APM
+		if (apm_resume) {
+			apm_resume = 0;
+			start_hc(uhci);
+			continue;
+		}
+#endif
 		uhci_check_configuration(uhci);
 
 		if(signal_pending(current)) {
@@ -877,9 +1046,11 @@ static int uhci_control_thread(void * __uhci)
 		}
 	}
 
+#if 0
 	if(uhci->root_hub)
 		for(i = 0; i < uhci->root_hub->usb->maxchild; i++)
 			usb_disconnect(uhci->root_hub->usb->children + i);
+#endif
 
 	reset_hc(uhci);
 	release_region(uhci->io_addr, 32);
@@ -947,6 +1118,37 @@ static int init_uhci(struct pci_dev *dev)
 	return -1;
 }
 
+#ifdef CONFIG_APM
+static int handle_apm_event(apm_event_t event)
+{
+	static int down = 0;
+
+	switch (event) {
+	case APM_SYS_SUSPEND:
+	case APM_USER_SUSPEND:
+		if (down) {
+			printk(KERN_DEBUG "uhci: received extra suspend event\n");
+			break;
+		}
+		down = 1;
+		break;
+	case APM_NORMAL_RESUME:
+	case APM_CRITICAL_RESUME:
+		if (!down) {
+			printk(KERN_DEBUG "uhci: received bogus resume event\n");
+			break;
+		}
+		down = 0;
+		if (waitqueue_active(&uhci_configure)) {
+			apm_resume = 1;
+			wake_up(&uhci_configure);
+		}
+		break;
+	}
+	return 0;
+}
+#endif
+
 int init_module(void)
 {
 	int retval;
@@ -970,8 +1172,19 @@ int init_module(void)
 		usb_mouse_init();
 		usb_kbd_init();
 		hub_init();
+#ifdef CONFIG_APM
+		apm_register_callback(&handle_apm_event);
+#endif
 
 		return 0;
 	}
 	return retval;
 }
+
+void cleanup_module(void)
+{
+#ifdef CONFIG_APM
+	apm_unregister_callback(&handle_apm_event);
+#endif
+}
+
