@@ -475,14 +475,66 @@ int usb_get_configuration(struct usb_device *dev)
  */
 void usb_new_device(struct usb_device *dev)
 {
+	int addr, i;
+
 	printk("USB new device connect, assigned device number %d\n",
 		dev->devnum);
-	if (usb_set_address(dev))
+
+	dev->maxpacketsize = 0;		/* Default to 8 byte max packet size */
+
+	addr = dev->devnum;
+	dev->devnum = 0;
+
+	/* Slow devices */
+	for (i = 0; i < 5; i++) {
+		if (!usb_get_descriptor(dev, USB_DT_DEVICE, 0, &dev->descriptor, 8))
+			break;
+
+		printk("get_descriptor failed, waiting\n");
+		wait_ms(200);
+	}
+	if (i == 5) {
+		printk("giving up\n");
 		return;
-	if (usb_get_device_descriptor(dev))
+	}
+
+#if 0
+	printk("maxpacketsize: %d\n", dev->descriptor.bMaxPacketSize0);
+#endif
+	switch (dev->descriptor.bMaxPacketSize0) {
+		case 8: dev->maxpacketsize = 0; break;
+		case 16: dev->maxpacketsize = 1; break;
+		case 32: dev->maxpacketsize = 2; break;
+		case 64: dev->maxpacketsize = 3; break;
+	}
+#if 0
+	printk("dev->mps: %d\n", dev->maxpacketsize);
+#endif
+
+	dev->devnum = addr;
+
+	if (usb_set_address(dev)) {
+		printk("Unable to set address\n");
+		/* FIXME: We should disable the port */
 		return;
-	if (usb_get_configuration(dev))
+	}
+
+	wait_ms(10);	/* Let the SET_ADDRESS settle */
+
+	if (usb_get_device_descriptor(dev)) {
+		printk("Unable to get device descriptor\n");
 		return;
+	}
+
+	if (usb_get_configuration(dev)) {
+		printk("Unable to get configuration\n");
+		return;
+	}
+
+#if 0
+	printk("Vendor: %X\n", dev->descriptor.idVendor);
+	printk("Product: %X\n", dev->descriptor.idProduct);
+#endif
 
 	usb_device_descriptor(dev);
 }
