@@ -3,6 +3,14 @@
 
 #include <linux/types.h>
 #include <linux/list.h>
+#include <linux/sched.h>
+
+static __inline__ void wait_ms(unsigned int ms)
+{
+        current->state = TASK_UNINTERRUPTIBLE;
+        schedule_timeout(1 + ms / 10);
+}
+
 
 typedef struct {
   unsigned char requesttype;
@@ -215,6 +223,8 @@ struct usb_bus {
 struct usb_device {
 	int devnum;						/* Device number on USB bus */
 	int slow;						/* Slow device? */
+	int maxpacketsize;					/* Maximum packet size */
+
 	struct usb_bus *bus;					/* Bus we're apart of */
 	struct usb_driver *driver;				/* Driver */
 	struct usb_device_descriptor descriptor;		/* Descriptor */
@@ -300,7 +310,7 @@ extern int  usb_parse_configuration(struct usb_device *dev, void *buf, int len);
 
 static inline unsigned int __create_pipe(struct usb_device *dev, unsigned int endpoint)
 {
-	return (dev->devnum << 8) | (endpoint << 15) | (dev->slow << 26);
+	return (dev->devnum << 8) | (endpoint << 15) | (dev->slow << 26) | dev->maxpacketsize;
 }
 
 static inline unsigned int __default_pipe(struct usb_device *dev)
